unit Process;

interface

uses Classes;

type
  // forward definition
  TExecuter = class;

  TForkerEventProc = procedure of Object;
  TforkerWhichProc = (wpParent, wpChild);

  TForker = class
  private
    FDebug        : boolean;
    FOnChild      : TForkerEventProc;
    FWait         : boolean;
    FExec         : TExecuter;
    FStatus       : integer;

    procedure DebugMsg(msg : string);
  public
    constructor Create;
    function    DoFork : TForkerWhichProc;
    procedure   DoWait;

  published
    property Debug        : boolean          read FDebug   write FDebug;
    property OnChild      : TForkerEventProc read FOnChild write FOnChild;
    property WaitForChild : boolean          read FWait    write FWait;
    property Exec         : TExecuter        read FExec    write FExec;
    property WaitStatus   : integer          read FStatus;

  end;

  TArgArray = array of AnsiString;

  TExecuter = class
  private
    FDebug    : Boolean;
    FEnv      : TStrings;
    FParms    : TStrings;
    FProcName : AnsiString;

    function StringListToCarray( cmd : AnsiString;
                                  strlst : TStrings ) : TArgArray;
    procedure DebugMsg(msg : string);
  protected
    function ListArgArray(aa : TArgArray) : string;

  public
    constructor Create;
    destructor  Destroy; override;
    procedure   Exec;

  published
    property Debug       : boolean      read FDebug    write FDebug;
    property Environment : TStrings     read FEnv      write Fenv;
    property Parameters  : TStrings     read FParms    write FParms;
    property ProcessName : AnsiString   read FProcName write FProcName;
  end;


implementation

uses
  Libc,SysUtils;

constructor TExecuter.Create;
begin
  inherited Create;
  FDebug    := false;
  FProcName := '';
  FEnv      := TStringList.Create;
  FParms    := TStringList.Create;
end;

destructor TExecuter.Destroy;
begin
  FEnv.Free;
  FParms.Free;
  inherited Destroy;
end;

procedure TExecuter.DebugMsg(msg : string);
begin
  if FDebug then
    writeln(msg);
end;

function TExecuter.StringListToCarray( cmd : AnsiString;
                                       strlst : TStrings ) : TArgArray;

var
  i,cnt : integer;

begin
  // set the array one bigger to account for the "NULL" end of array terminator

  cnt := strlst.Count+1;
  if cmd <> '' then
    inc(cnt);

  DebugMsg('Setting Results length.. cnt is ' + IntToStr(cnt));
  SetLength(Result, cnt);

  DebugMsg('Before Result[0] = cmd...');
  // when cmd is nothing, this will be overwritten
  Result[0] := cmd;

  DebugMsg('Before for loop...');

  for i:= 0 to strlst.Count-1 do
  begin
    Result[i+1] := strlst.Strings[i];
  end;
end;

procedure TExecuter.Exec;
var
  env   : TArgArray;
  parms : TArgArray;
  cmd   : AnsiString;
  j     : integer;

begin
  cmd   := '';
  DebugMsg('Environment vars ' + FEnv.Text);
  env   := StringListToCarray(cmd,FEnv);
  DebugMsg('EnvArgArray is ' + ListArgArray(env));

  cmd   := FProcName;
  DebugMsg('Parameters are ' + FParms.Text);
  parms := StringListToCarray(cmd,FParms);
  DebugMsg('ArgArray is ' + ListArgArray(parms));

  DebugMsg('Before execv cmd is ' + cmd + ' parms are ' + FParms.Text);
  j := execv(PChar(cmd), PPChar(@parms[0]));
  if j = -1 then
    raise Exception.CreateFmt('execv failed error %d',[GetLastError]);
  // when properly executed, execv will never return...
end;

function TExecuter.ListArgArray(aa : TArgArray) : string;
var
  i,len : integer;

begin
  len    := length(aa);
  Result := 'Length of array: ' + IntToStr(len) + ': ';
  for i:=0 to len-1 do
  begin
    Result := Result + IntToStr(i) + ') ' + aa[i] + ' |';
  end;
end;

constructor TForker.Create;
begin
  inherited Create;
  FDebug := false;
  FWait  := true;
end;

function   TForker.DoFork : TForkerWhichProc;
var
  i : integer;

begin
  i := fork;
  if i = -1 then
  begin
    // uh oh error!
    raise Exception.CreateFmt('Unable to fork: Error is %d',[GetLastError]);
  end
  else if i = 0 then
  begin
    // we are in the child...
    Result := wpChild;

    // call the child
    if Assigned (FOnChild) then
      FOnChild
    else if Assigned (FExec) then
    begin
      // do the exec thing..
      DebugMsg('Before Exec..');
      FExec.Exec;
    end;

    // otherwise we fall through and let the
    // caller handle it..
  end
  else
  begin
    // we are the parent...
    Result := wpParent;

    if FWait then
    begin
       // wait for child
       wait(@FStatus);
    end;
  end;
end;

procedure TForker.DoWait;
begin
  if not FWait then
    wait(@FStatus);
end;


procedure TForker.DebugMsg(msg : string);
begin
  if FDebug then
    writeln(msg);
end;

end.
