program simpleshell;

(**
 * program simpleshell;
 *
 * Description: Demonstrates the how a command shell works
 *)

uses Libc;

type
  TArgArray = array of AnsiString;

var
  i,j,status  : integer;
  cmdline,cmd : string;
  args        : TArgArray;

(**
 * procedure ParseArgs(commandLine : string; var cmd : string; 
 *                     var args : TArgArray);
 * 
 * Description: Parses the command line into a command and it's arguments
 *
 * Notes: Designed for use with exec* functions. The argument list starts
 *        with the command, followed by the arguments, if any.  These 
 *        functions expect the ArgArray to be sized one (1) bigger than 
 *        the actual number of arguments. This is because the last 
 *        argument needs to be nil to indicate the end of the argument list.
 *
 *        This function DOES NOT handle quoted arguments.
 *)

  procedure ParseArgs(commandline : string; var cmd : string; 
                      var args : TArgArray);

  var
    tmp        : string;
    i          : integer;
    inspos     : integer;
    curLen     : integer;

  begin
    // assume there are no arguments..
    insPos := 0;
    curLen := 1;
    SetLength(args,curLen);
   
    tmp        := commandline;
    i          := Pos(' ',tmp);
    while (i > 0) do
    begin
       // found an argument
       inc(curLen);
       SetLength(args,curLen); 

       // extract the string from 1 to i-1
       args[insPos] := Copy(tmp,1,(i-1));

       tmp := Copy(tmp,i+1,length(tmp));
       i := Pos(' ',tmp);
       inc(insPos);
    end;

    if (tmp <> '') then
    begin
       // found an argument
      inc(curLen);
      SetLength(args,curLen); 
 
      // extract the string from 1 to i-1
      args[insPos] := Copy(tmp,1,length(tmp));
    end;
 
    // now get the command from the first argument
    cmd := args[0];

{$IFDEF DEBUG}
    writeln('The length of args is ',length(args));
    writeln('Parsed, cmd is ',cmd);
    for i := 0 to length(args)-1 do
    begin
       writeln('Arg[',i,'] is ',args[i]);
    end;
    writeln('-----------------'); 
{$ENDIF}
  end;

begin
  while true do
  begin
     write('[ss]$ ');
     readln(cmdline);
     if (cmdline = '') then
       break;

     // search for arguments in the string
     ParseArgs(cmdline,cmd,args);

     i := fork;
     if (i = -1) then
     begin
       // an error occured report it and exit the loop
       perror(PChar('Error attempting to fork..'));
       break;
     end
     else if (i = 0) then
     begin
       // in the new, child process so execute the command..
       j := execv(PChar(cmd), PPChar(@args[0]));       
       if (j = -1) then
       begin
         // an error occured report it and exit the loop
         writeln('Error executing ',cmd,' cmdline ',
                 cmdline,' the error was ',GetLastError);
         perror(PChar('Error: '));
         break;
       end;
     end
     else
     begin
       // in the parent process wait for my child to exit
       wait(@status);
       if (status <> 0) then
       begin
          // an error occured report it and exit the loop
          perror(PChar('Error waiting for child process'));
          break;
       end;
     end;
  end;
end.
